import util/preprocessing
import util/matrix
import util/feature_extraction/text
import nn/network


import strformat
import sequtils
import json
import math
import times


proc loadData: tuple[corpus, results, testset, testResults: Matrix[string]] =
    # Loads the dataset and the testset for the supervised learning test
    let dataset = parseFile("assets/datasets/spamham.json")
    let testset = parseFile("assets/testsets/spamham.json")
    var corpus: seq[string] = @[]
    var results: seq[string] = @[]
    var test: seq[string] = @[]
    var testResults: seq[string] = @[]
    for label in dataset.keys():
        for sentence in dataset[label]:
            corpus.add(sentence.getStr())
            results.add(label)
    for label in testset.keys():
        for sentence in testset[label]:
            test.add(sentence.getStr())
            testResults.add(label)
    result = (corpus: newMatrix(corpus),
              results: newMatrix(results),
              testset: newMatrix(test), 
              testResults: newMatrix(testResults))


proc testMetrics(predictions, y: Matrix[int]): tuple[accuracy, precision, recall: float, tP, tN, fP, fN: int] =
    # Computes the test metrics given the predictions and the 
    # expected results
    var confusion = @[0, 0, 0, 0] # TP, TN, FP, FN
    var success = 0
    var i = 0
    while i < y.len():
        inc(i)
        var predicted = predictions[0][i]
        var expected = y[0][i]
        if predicted == expected:
            success += 1
            if predicted == 1:
                confusion[0] += 1
            else:
                confusion[1] += 1
        elif predicted == 1 and expected == 0:
            confusion[2] += 1
        else:
            confusion[3] += 1
    result = (accuracy: success / len(predictions),
              precision: confusion[0] / (confusion[0] + confusion[2]),
              recall: confusion[0] / (confusion[0] + confusion[3]),
              tP: confusion[0], 
              tN: confusion[1], 
              fP: confusion[2],
              fN: confusion[3])


proc main =
    # Runs the test
    const stopwords = @["i", "me", "my", "myself", "we", "our", "ours", 
 "ourselves", "you", "you're", "you've", "you'll", "you'd",
 "your", "yours", "yourself", "yourselves", "he", "him", "his", 
 "himself", "she", "she's", "her", "hers", "herself", "it", "it's",
 "its", "itself", "they", "them", "their", "theirs", "themselves", "what", 
 "which", "who", "whom", "this", "that", "that'll", "these", "those", "am", 
 "is", "are", "was", "were", "be", "been", "being", "have", "has", "had", 
 "having", "do", "does", "did", "doing", "a", "an", "the", "and", "but", 
 "if", "or", "because", "as", "until", "while", "of", "at", "by", "for", 
 "with", "about", "against", "between", "into", "through", "during", "before", 
 "after", "above", "below", "to", "from", "up", "down", "in", "out", "on", 
 "off", "over", "under", "again", "further", "then", "once", "here", "there", 
 "when", "where", "why", "how", "all", "any", "both", "each", "few", "more", 
 "most", "other", "some", "such", "no", "nor", "not", "only", "own", "same", 
 "so", "than", "too", "very", "s", "t", "can", "will", "just", "don", "don't",
 "should", "should've", "now", "d", "ll", "m", "o", "re", "ve", "y", "ain", 
 "aren", "aren't", "couldn", "couldn't", "didn", "didn't", "doesn", 
 "doesn't", "hadn", "hadn't", "hasn", "hasn't", "haven", "haven't", 
 "isn", "isn't", "ma", "mightn", "mightn't", "mustn", "mustn't", "needn", 
 "needn't", "shan", "shan't", "shouldn",  "shouldn't", "wasn", "wasn't", 
 "weren", "weren't", "won", "won't", "wouldn", "wouldn't"]
    const epochs = 100
    const batch = 400
    const inputSize = 512
    let encoder = newLabelEncoder()
    let cleaner = newTextPreprocessor(stopwords=newMatrix(stopwords), toLower=true,
                                      stripPunctuation=true, normalize=false)
    let vectorizer = newTFIDFVectorizer(minDf=0.01, maxDf=0.7, preprocessor=cleaner,
                                        sublinearTf=false, smoothIdf=true)
    var classifier = newNeuralNetwork(@[newDenseLayer(inputSize, 8),
                                        newDenseLayer(8, 16),
                                        newDenseLayer(16, 2)],
                                        lossFunc=MSE,
                                        activationFunc=SiLU,
                                        learnRate=0.001,
                                        momentum=0.0,
                                        weightRange=(-1.0, 1.0),
                                        biasRange=(-1.0, 1.0))
    echo "ProjectSydney v0.2b - Accuracy test"
    echo "Loading dataset and testset"
    let loadTime = cpuTime()
    let data = loadData()
    echo &"Data loaded in {cpuTime() - loadTime:.2f} seconds"
    echo "Processing and vectorizing dataset with TF-IDF weigths"
    let vectorTime = cpuTime()
    let xTrain = vectorizer.fitTransform(data.corpus, data.corpus)
    let yTrain = encoder.fitTransform(data.results, data.results)[0]
    echo &"Data vectorized in {cpuTime() - vectorTime:.2f} seconds"
    echo &"\t- Feature count: {len(vectorizer.getFeatureNames())}"
    echo &"\t- Vocabulary size: {len(vectorizer.getVocabulary())}"
    echo &"\t- Corpus size: {len(data.corpus)}"
    let yTest = encoder.transform(data.testResults)[0]
    let xTest = vectorizer.transform(data.testset)
    var tempData: seq[float] = newSeqOfCap[float](inputSize)
    var trainData: seq[tuple[x, y: Matrix[float]]] = @[]
    var testData: seq[tuple[x, y: Matrix[float]]] = @[]
    var testTruth: seq[int] = @[]
    # Pad the data to fit into the network
    for i, row in xTrain:
        for e in row:
            if tempData.len() == inputSize:
                break
            tempData.add(e)
        while tempData.len() < inputSize:
            tempData.add(0.0)
        if yTrain[i] == 1:
            trainData.add((newMatrix[float](tempData), newMatrix[float](@[1.0, 0.0])))
        else:
            trainData.add((newMatrix[float](tempData), newMatrix[float](@[0.0, 1.0])))
        tempData.setLen(0)
    for i, row in xTest:
        for e in row:
            if tempData.len() == inputSize:
                break
            tempData.add(e)
        while tempData.len() < inputSize:
            tempData.add(0.0)
        if yTest[i] == 1:
            testTruth.add(1)
            testData.add((newMatrix[float](tempData), newMatrix[float](@[1.0, 0.0])))
        else:
            testTruth.add(0)
            testData.add((newMatrix[float](tempData), newMatrix[float](@[0.0, 1.0])))
        tempData.setLen(0)
    echo "Classifier parameters"
    echo &"\t- Learn rate: {classifier.learnRate}"
    echo &"\t- Momentum: {classifier.momentum}"
    stdout.write("\t- Network layout: ")
    for i, layer in classifier.layers:
        stdout.write(&"{layer.inputSize}x{layer.outputSize}")
        if i < classifier.layers.high():
            stdout.write(" -> ")
    echo ""
    echo &"Training neural network for {epochs} epochs with batch size of {batch}"
    let trainTime = cpuTime()
    classifier.train(epochs, batch, trainData, testData)
    echo &"Training completed in {cpuTime() - trainTime:.2f} seconds, running test"
    echo "Test parameters"
    echo &"\t- Test size: {len(data.testset)}"
    let testTime = cpuTime()
    var pred: seq[int] = @[]
    for sample in xTest:
        pred.add(classifier.predict(sample.copy()))
    let predictions = newMatrix[int](pred)
    let metrics = testMetrics(predictions, newMatrix[int](testTruth))
    echo &"Test completed in {cpuTime() - testTime:.2f} seconds, metrics below"
    echo &"\t- Accuracy: {metrics.accuracy * 100:.2f}%"
    echo &"\t- Recall: {metrics.recall:.2f}"
    echo &"\t- Precision: {metrics.precision:.2f}"
    echo &"\t- F1-score: {pow((pow(metrics.precision, -1) + pow(metrics.recall, -1)) / 2, -1):.2f}"
    echo "\tConfusion matrix"
    echo &"\t\t- True positives: {metrics.tP}"
    echo &"\t\t- True negatives: {metrics.tN}"
    echo &"\t\t- False negatives: {metrics.fN}"
    echo &"\t\t- False positives: {metrics.fP}"



when isMainModule:
    setControlCHook(proc () {.noconv.} = quit(0))
    main()


